'use strict';

const showLog = false;
function log() {
    showLog && console.log.apply(console, arguments);
}

if (chrome.extension.inIncognitoContext) {

    getSavedOptions().then(options => {
        log('saved opts', options);
        if (options.aggressiveModeEnabled) {
            chrome.runtime.sendMessage({clearPrivateData: true, callFromContentScript: true});
        }
    });

    // This will get the source code of the function definition.
    const code = getFunctionSourceCodeWithSelfCall(runScriptsInWebpage);

    // This will synchronously execute that code.
    // We do this because we need to run our code before the dom in the webpage gets built.
    // If we tried to instead inject a <script src=foo.js>
    // it would a_synchronously execute much later, which wouldn't work very effectively because
    // other inline scripts in the webpage would run before our code.
    injectScriptTextViaTextNode(code).then(() => {
        // log('inj incog');
    });

}


/**
 * Mimics a script like <script>...code here...</script>
 *
 * This is synchronous, so calling this will immediately execute the
 * code as soon as you call this func, before the promise even has a chance to resolve.
 *
 * This is in contrast with injecting a script tag with
 * a url + src, which will not synchronously execute the code.
 *
 * @param code
 * @returns {Promise<unknown>}
 */
function injectScriptTextViaTextNode(code) {
    return new Promise(function (resolve, reject) {
        const elem = document.createElement("script");
        elem.type = "text/javascript";

        elem.onload = resolve;
        elem.onerror = reject;

        elem.appendChild(document.createTextNode(code));
        (document.head || document.documentElement).appendChild(elem);
    });
}

function getSavedOptions() {
    const aggressiveModeEnabledDefaultValue = false;
    return new Promise(resolve => {
        chrome.storage.sync.get({
            "aggressiveModeEnabled": aggressiveModeEnabledDefaultValue,
        }, resolve);
    });
}

function runScriptsInWebpage() {
    "use strict";

    /**
     * Returns the greater of your argument, or 120MB + 1
     *
     * @param {number} actualGrantedQuotaBytes
     * @returns {number}
     */
    function getDeceptiveStorageQuota(actualGrantedQuotaBytes) {
        // The incognito detection method used by some websites assumes that the max storage quota in incognito is 120MB, so we will use 120MB + 1 single more byte.
        // If the Times were to see this code, me using exactly 1 more byte would be more taunting and likely to upset them, which would make me smile lol.
        // I think the real number is actually 120 * (2 ** 20) but some incognito detection scripts just use the
        // number 120000000, so we will use the bigger number as it will hopefully work better and fool more sites.
        // ref: https://bugs.chromium.org/p/chromium/issues/detail?id=800329
        const nonIncognitoMinimumStorageBytes = 120 * (2 ** 20);
        const fakeQuotaBytes = nonIncognitoMinimumStorageBytes + 1;

        // We only change the quota if it's too small. Otherwise we leave it alone so it's accurate.
        return Math.max(actualGrantedQuotaBytes, fakeQuotaBytes);
    }

    function mockRequestFileSystemFn(type, size, successCallback, errorCallback) {
        setTimeout(() => {
            // Call the callback w/ a poor imitation of a FilesystemObject.
            const mockFilesystemObject = {};
            successCallback(mockFilesystemObject);
        }, 0);
    }

    function mockStorageEstimateQuotaFn() {
        // Return a promise that tries hard to mimic the native storage.estimate() promise, forwarding arguments.
        return new Promise((resolve, reject) => {
            mockStorageEstimateQuotaFn.nativeFn.call(navigator.storage).then(function(estimateDetails) {

                // We will use the native object, and only shadow the quota property to overwrite it with our larger quota.
                const newDeceptiveEstimate = Object.create(estimateDetails);
                newDeceptiveEstimate.quota = getDeceptiveStorageQuota(estimateDetails.quota);

                // There should only be 1 argument, but we'll forward all args for future-proofing and max compatibility.
                arguments[0] = newDeceptiveEstimate;
                resolve(...arguments);
            }, reject);
        });
    }

    function mockWindowWebkitStorageInfoFn(temporaryOrPersistent, callback) {
        // We only mock calls for temp storage.
        if (temporaryOrPersistent !== webkitStorageInfo.TEMPORARY) {
            // Probably a call for persistent storage. We won't mock.
            return webkitStorageInfo.queryUsageAndQuota(temporaryOrPersistent, callback);
        }

        return mockWindowWebkitStorageInfoFn.nativeFn.call(webkitStorageInfo, function(usedBytes, grantedQuotaBytes) {
            // There should only be 2 arguments, but we'll forward all args for future-proofing and max compatibility.
            arguments[1] = getDeceptiveStorageQuota(grantedQuotaBytes);
            callback(...arguments);
        });
    }

    function mockNavigatorWebkitTemporaryStorageEstimateQuotaFn(successCallback, errorCallback) {
        return mockNavigatorWebkitTemporaryStorageEstimateQuotaFn.nativeFn.call(
            navigator.webkitTemporaryStorage,
            function(usedBytes, grantedQuotaBytes) {
                // There should only be 2 arguments, but we'll forward all args for future-proofing and max compatibility.
                arguments[1] = getDeceptiveStorageQuota(grantedQuotaBytes);
                successCallback(...arguments);
            },
            errorCallback
        );
    }


    if (window.RequestFileSystem) {
        window.RequestFileSystem = mockRequestFileSystemFn;
    }

    if (window.webkitRequestFileSystem) {
        window.webkitRequestFileSystem = mockRequestFileSystemFn;
    }

    // We also mock the StorageManager.estimate() method because now some sites are using this as a new incognito detection method
    // which looks at the storage quota to determine if incog.
    // Infoz https://mishravikas.com/articles/2019-07/bypassing-anti-incognito-detection-google-chrome.html
    if (navigator.storage && navigator.storage.estimate) {
        // Store a ref to orig func before we overwrite it w/ our new shim / mock.
        mockStorageEstimateQuotaFn.nativeFn = navigator.storage.estimate;
        navigator.storage.estimate = mockStorageEstimateQuotaFn;
    }

    // webkitTemporaryStorage is deprecated and/or webkit specific. But, we still mock it because some sites could use it.
    if (navigator.webkitTemporaryStorage && navigator.webkitTemporaryStorage.queryUsageAndQuota) {
        // Store a ref to orig func before we overwrite it w/ our new shim / mock.
        mockNavigatorWebkitTemporaryStorageEstimateQuotaFn.nativeFn = navigator.webkitTemporaryStorage.queryUsageAndQuota;
        navigator.webkitTemporaryStorage.queryUsageAndQuota = mockNavigatorWebkitTemporaryStorageEstimateQuotaFn;
    }

    // webkitStorageInfo is deprecated and/or webkit specific. But, we still mock it because some sites could use it.
    if (window.webkitStorageInfo && window.webkitStorageInfo.queryUsageAndQuota) {
        // Store a ref to orig func before we overwrite it w/ our new shim / mock.
        mockWindowWebkitStorageInfoFn.nativeFn = window.webkitStorageInfo.queryUsageAndQuota;
        window.webkitStorageInfo.queryUsageAndQuota = mockWindowWebkitStorageInfoFn;
    }

}

/**
 * Gets the js source code for a function, and also generates some
 * code that will call the function. This is useful with eval(), or when dynamically creating
 * and executing script elements.
 *
 * @example
 * // Assuming you have the following function definition:
 * function sayHello() { alert('Hello') }
 *
 * // The following assertion will be true, as an example:
 * // "(function sayHello() { alert('Hello') })();" === getFunctionSourceCodeWithSelfCall(sayHello);
 *
 * @param {function} fn - a function that MUST have a name property.
 * @returns {string} - js code of the function
 */
function getFunctionSourceCodeWithSelfCall(fn) {
    return `(${fn})();`; // A function's toString() method handily returns us the source code.
}